{+--------------------------------------------------------------------------+
 | Class:       TmTSearchOP  ( TurboSearch )
 | Created:     8.97
 | Author:      Martin Waldenburg
 | Description: A very fast search engine, about twice as fast as Boyer-Moore,
 |              based on an article in the German magazine c't (8/97).
 |              The original is in 'C '. www.Heise.com.
 |              You can search case sensitive or case insensitive.
 |              Look_At is implemented.
 | Version:     2.2
 | Copyright (c) 1997 - 1998 Martin Waldenburg
 | All rights reserved.
 |
 | LICENCE CONDITIONS
 |
 | USE OF THE ENCLOSED SOFTWARE
 | INDICATES YOUR ASSENT TO THE
 | FOLLOWING LICENCE CONDITIONS.
 |
 |
 |
 | These Licence Conditions are exlusively
 | governed by the Law and Rules of the
 | Federal Republic of Germany.
 |
 | Redistribution and use in source and binary form, with or without
 | modification, are permitted provided that the following conditions
 | are met:
 |
 | 1. Software using this code must contain a visible line of credit,
 |    to Martin Waldenburg and the German magazine c't (8/97).
 |
 | DISCLAIMER:
 |
 | THIS SOFTWARE IS PROVIDED BY THE AUTHOR 'AS IS'.
 |
 | ALL EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 | THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 | PARTICULAR PURPOSE ARE DISCLAIMED.
 |
 | IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 | INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 | (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 | OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 | INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 | WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 | NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 | THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 |
 |  Martin.Waldenburg@T-Online.de

+--------------------------------------------------------------------------+}
unit mwTSearchOP;


interface

uses
  Windows,
  SysUtils,
  Messages,
  Classes,
  Graphics,
  Controls,
  Forms,
  Dialogs,
  Menus;

type
  TmTSearchOP = class(Tobject)
  private
    Run: PChar;
    Origin: Pchar;
    TheEnd: PChar;
    Pat: String;
    fCount: Integer;
    Look_At: Integer;
    SimpleJump: Integer;
    PatLen, PatLenPlus: Integer;
    Shift: array[0..255] of Integer;
    fFinished: Boolean;
    fSensitive: Boolean;
    fWhole: Boolean;
    function FindFirstInSensitive(NewOrigin: PChar; NewLength: LongInt): Integer;
    function FindFirstSensitive(NewOrigin: PChar; NewLength: LongInt): Integer;
    function FindFirstWhole(NewOrigin: PChar; NewLength: LongInt): Integer;
    function GetFinished: Boolean;
    procedure InitInSensitive(NewPattern: String);
    procedure InitSensitive(NewPattern: String);
    function NextInSensitive: Integer;
    function NextSensitive: Integer;
    function NextWhole: Integer;
    procedure SetPattern(const Value: String);
    procedure SetSensitive(const Value: Boolean);
    function IsWhole: Boolean;
  protected
  public
    constructor Create;
    destructor Destroy; override;
    function FindFirst(NewOrigin: PChar; NewLength: LongInt): Integer;
    function Next: Integer;
    property Count: Integer read fCount write fCount;
    property Finished: Boolean read GetFinished;
    property Pattern: String read Pat write SetPattern;
    property Sensitive: Boolean read fSensitive write SetSensitive;
    property Whole: Boolean read fWhole write fWhole;
  published
  end;

var
  CompTable: array[#0..#255] of byte;

implementation

constructor TmTSearchOP.Create;
begin
  inherited Create;
  Pat := '';
  PatLen := 0;
  Run := nil;
  TheEnd := nil;
  Run := nil;
  fCount := 0;
  fSensitive := True;
  fWhole := False;
end; { Create }

procedure MakeCompTable;
var
  I: Byte;
begin
  for I := 0 to 255 do
    CompTable[Char(I)] := ord(AnsiLowerCase(Char(I))[1]);
end; { MakeCompTable }

function TmTSearchOP.GetFinished: Boolean;
begin
  fFinished := False;
  if Run >= TheEnd then fFinished := True;
  if PatLen > (TheEnd - Origin) then fFinished := True;
  Result := fFinished;
end; { GetFinished }

procedure TmTSearchOP.InitSensitive(NewPattern: String);
var
  I: Byte;
begin
  Pat := NewPattern;
  PatLen := Length(Pat);
  PatLenPlus := PatLen + 1;
  Look_At := 1;
  for I := 0 to 255 do Shift[I] := PatLenPlus;
  for I := 1 to PatLen do Shift[ord(Pat[I])] := PatLenPlus - I;
  while Look_at < PatLen do
  begin
    if Pat[PatLen] = Pat[PatLen - (Look_at)] then exit;
    inc(Look_at);
  end;
end; { Init }

function TmTSearchOP.FindFirstSensitive(NewOrigin: PChar; NewLength: LongInt): Integer;
begin
  Origin := NewOrigin;
  TheEnd := NewOrigin + NewLength;
  Result := -1;
  fCount := 0;
  if (TheEnd - Origin) >= PatLen then
  begin
    Run := (NewOrigin - 1);
    Result := NextSensitive;
  end;
end; { FindFirst }

function TmTSearchOP.NextSensitive: Integer;
var
  I: Integer;
  J: PChar;
begin
  Result := -1;
  inc(Run, PatLen);
  if Finished then Exit;
  while Run < TheEnd do
  begin
    if Pat[Patlen] <> Run^ then inc(Run, Shift[ord((Run + 1)^)])
    else
    begin
      J := Run - PatLen + 1;
      I := 1;
      while Pat[I] = J^ do
      begin
        inc(I);
        if I = PatLen then
        begin
          inc(fCount);
          Result := Run - Origin - Patlen + 1;
          exit;
        end;
        inc(J);
      end;
      inc(Run, Look_at + Shift[ord((Run + Look_at)^)]-1); // Suggestion of Angus Johnson
    end;
  end;
end; { Next }

procedure TmTSearchOP.InitInSensitive(NewPattern: String);
var
  I: Byte;
begin
  Pat := NewPattern;
  PatLen := Length(Pat);
  PatLenPlus := PatLen + 1;
  Look_At := 1;
  for I := 0 to 255 do Shift[I] := PatLenPlus;
  for I := 1 to PatLen do Shift[CompTable[Pat[i]]] := PatLenPlus - I;
  while Look_at < PatLen do
  begin
    if CompTable[Pat[PatLen]] = CompTable[Pat[PatLen - (Look_at)]] then exit;
    inc(Look_at);
  end;
end; { IInit }

function TmTSearchOP.FindFirstInSensitive(NewOrigin: PChar; NewLength: LongInt): Integer;
begin
  Origin := NewOrigin;
  TheEnd := NewOrigin + NewLength;
  Result := -1;
  fCount := 0;
  if (TheEnd - Origin) >= PatLen then
  begin
    Run := (NewOrigin - 1);
    Result := NextInSensitive;
  end;
end; { IFindFirst }

function TmTSearchOP.NextInSensitive: Integer;
var
  I: Integer;
  J: PChar;
begin
  Result := -1;
  inc(Run, PatLen);
  if Finished then Exit;
  while Run < TheEnd do
  begin
    if CompTable[Pat[Patlen]] <> CompTable[Run^] then
      inc(Run, Shift[CompTable[(Run + 1)^]])
    else
    begin
      J := Run - PatLen + 1;
      I := 1;
      while CompTable[Pat[I]] = CompTable[J^] do
      begin
        inc(I);
        if I = PatLen then
        begin
          inc(fCount);
          Result := Run - Origin - Patlen + 1;
          exit;
        end;
        inc(J);
      end;
      inc(Run, Look_At + Shift[CompTable[(Run + Look_at)^]]-1); // Suggestion of Angus Johnson
    end;
  end;
end; { INext }

destructor TmTSearchOP.Destroy;
begin
  inherited Destroy;
end; { Destroy }

procedure TmTSearchOP.SetPattern(const Value: String);
begin
  Case fSensitive of
    True: InitSensitive(Value);
    False: InitInSensitive(Value);
  end;
end;

procedure TmTSearchOP.SetSensitive(const Value: Boolean);
begin
  fSensitive := Value;
  if Pat <> '' then
    Case fSensitive of
      True: InitSensitive(Pat);
      False: InitInSensitive(Pat);
    end;
end;

function TmTSearchOP.FindFirst(NewOrigin: PChar;
  NewLength: Integer): Integer;
begin
  Case fWhole of
    True: Result:= FindFirstWhole(NewOrigin, NewLength);
    False:
      Case fSensitive of
        True: Result := FindFirstSensitive(NewOrigin, NewLength);
        False: Result := FindFirstInSensitive(NewOrigin, NewLength);
      end;
  end;
end;

function TmTSearchOP.Next: Integer;
begin
  Case fWhole of
    True: Result:= NextWhole;
    False:
      Case fSensitive of
        True: Result := NextSensitive;
        False: Result := NextInSensitive;
      end;
  end;
end;

function TmTSearchOP.IsWhole: Boolean;
var
  Before, After: PChar;
begin
  Result := True;
  Before := Run - PatLen;
  After := Run + 1;
  if (Before >= Origin) and (After <= TheEnd) then
  begin
    if Before^ in ['A'..'Z', 'a'..'z'] then Result := False else
    begin
      if After < TheEnd then
        if After^ in ['A'..'Z', 'a'..'z'] then Result := False
    end;
  end else
    if Before = Origin - 1 then
    begin
      if After < TheEnd then
        if After^ in ['A'..'Z', 'a'..'z'] then Result := False
    end;
end;

function TmTSearchOP.FindFirstWhole(NewOrigin: PChar;
  NewLength: Integer): Integer;
begin
  Case fSensitive of
    True:
      begin
        Result := FindFirstSensitive(NewOrigin, NewLength);
        while Result <> -1 do
        begin
          if IsWhole then break;
          dec(fCount);
          Result := NextSensitive;
        end;
      end;
    False:
      begin
        Result := FindFirstInSensitive(NewOrigin, NewLength);
        while Result <> -1 do
        begin
          if IsWhole then break;
          dec(fCount);
          Result := NextInSensitive;
        end;
      end;
  end;
end;

function TmTSearchOP.NextWhole: Integer;
begin
  Case fSensitive of
    True:
      begin
        Result := NextSensitive;
        while Result <> -1 do
        begin
          if IsWhole then break;
          dec(fCount);
          Result := NextSensitive;
        end;
      end;
    False:
      begin
        Result := NextInSensitive;
        while Result <> -1 do
        begin
          if IsWhole then break;
          dec(fCount);
          Result := NextInSensitive;
        end;
      end;
  end;
end;

Initialization
  MakeCompTable;

end.

